import sys
from abc import ABC
from typing import Any, Callable, Dict, List, Tuple

import h5py
import numpy as np
import pandas as pd
from mpi4py import MPI

class Event:
    def __init__(self,
                 index: int,
                 event_id: np.ndarray,
                 data: Dict[str, pd.DataFrame] = {}):
        self.index = index
        self.event_id = event_id
        self.data = data.copy()

    @property
    def name(self):
        r, sr, evt = self.event_id
        return f'r{r}_sr{sr}_evt{evt}'

    def __setitem__(self, key: str, item: pd.DataFrame):
        if type(key) != str:
            raise Exception('Key must be a string!')
        if type(item) != pd.DataFrame:
            raise Exception('Value must be a pandas DataFrame!')
        self.data[key] = item

    def __getitem__(self, key: str):
        if type(key) != str:
            raise Exception('Key must be a string!')
        return self.data[key]

    def __str__(self):
        ret = f'event {self.event_id}\n'
        for group, df in self.data.items():
            ret += f'  {group} ({df.shape[0]} rows):\n'
            for key in df.keys():
                ret += f'    {key}\n'
        return ret

class File:
    def __init__(self, fname: str, parKey: str = "/event_table/event_id"):
        self._colmap = {
            "event_table": {
                "nu_dir": [ "nu_dir_x", "nu_dir_y", "nu_dir_z" ],
                "nu_vtx": [ "nu_vtx_x", "nu_vtx_y", "nu_vtx_z" ],
                "nu_vtx_corr": [ "nu_vtx_corr_x", "nu_vtx_corr_y", "nu_vtx_corr_z" ],
            },
            "particle_table": {
                "start_position": [ "start_position_x", "start_position_y", "start_position_z" ],
                "end_position": [ "end_position_x", "end_position_y", "end_position_z" ],
                "start_position_corr": [ "start_position_corr_x", "start_position_corr_y", "start_position_corr_z" ],
                "end_position_corr": [ "end_position_corr_x", "end_position_corr_y", "end_position_corr_z" ],
            },
            "spacepoint_table": {
                "hit_id": [ "hit_id_u", "hit_id_v", "hit_id_y" ],
                "position": [ "position_x", "position_y", "position_z" ],
            },
            "pandoraPrimary_table": {
                "vtx": [ "vtx_x", "vtx_y", "vtx_z" ],
            },
        }

        # open the input HDF5 file in parallel
        self._fd = h5py.File(fname, "r", driver='mpio', comm=MPI.COMM_WORLD)

        # check if data partitioning key datasets exists in the file
        if parKey not in self._fd.keys():
            raise Exception(f'Error: dataset {parKey} is not found in file {fname}!')

        # parse the name of data partitioning key
        import os.path
        self._parTable = os.path.dirname(parKey)
        # remove leading '/'
        if self._parTable[0] == '/': self._parTable = self._parTable[1:]

        # extract dataset names: partitioning key, seq, and seq_cnt
        self._par_name = os.path.basename(parKey)
        self._seq_name = self._par_name + ".seq"
        self._cnt_name = self._par_name + ".seq_cnt"

        # obtain metadata of dataset parKey, later the dataset will be read
        # into self._index as a numpy array in data_partition()
        self._index = self._fd.get(parKey)
        self._num_events = self._index.shape[0]

        # self._groups is a python list, each member is a 2-element list consisting
        # of a group name, and a python list of dataset names
        self._groups = []

        # a python dictionary storing a sequence-count dataset in each group, keys
        # are group names, values are the sequence-count dataset subarrays assigned
        # to this process
        self._seq_cnt = {}
        self._evt_seq = {}

        self._whole_seq_cnt = {}
        self._whole_seq     = {}

        self._use_seq_cnt = True

        # partition based on event amount of particle table (default)
        self._evt_part = 2

        # a python nested dictionary storing datasets of each group read from the
        # input file. keys of self._data are group names, values are python
        # dictionaries, each has names of dataset in that group as keys, and values
        # storing dataset subarrays
        self._data = {}

        # _starts: data partition start indeices of all processes
        # _counts: data cmount assigned to each process
        starts = None
        counts = None

        # starting array index of parKey assigned to this process
        self._my_start = -1

        # number of array elements of parKey assigned to this process
        self._my_count = -1

    def __del__(self):
        if hasattr(self, '_fd') and self._fd: self._fd.close()

    def __len__(self):
        # inquire the number of unique event IDs in the input file
        return self._num_events

    def __str__(self):
        ret = ""
        for k1 in self._fd.keys():
            ret += f"{k1}:\n"
            for k2 in self._fd[k1].keys():
                if self._seq_name in k2: continue
                ret += f"    {k2}\n"
        return ret

    def __getitem__(self, idx: int):
        """load a single event from file"""
        self.read_data(idx, 1)
        ret = self.build_evt(idx, 1)
        return ret[0] if len(ret) else None

    def check_shape0(self,
                     group: str,
                     keys: List[str] = []) -> None:
        # Check if shape[0] of all datasets in keys are of the same size
        shape0 = self._fd[group][keys[0]].shape[0]
        for k in keys[1:]:
            if k == self._cnt_name: continue # exception is seq_cnt dataset
            if shape0 != self._fd[group][k].shape[0]:
               raise Exception(f'Dataset "/{group}/{k}" shape[0]={self._fd[group][k].shape[0]} inconsistent with {keys[0]}.shape[0]={shape0}')

    def add_group(self,
                  group: str,
                  keys: List[str] = []) -> None:

        # if no keys specified, append all columns in HDF5 group
        if not keys:
            # retrieve all the dataset names of the group
            keys = list(self._fd[group].keys())
            # datasets seq and seq_cnt are not needed
            if group != self._parTable and self._par_name in keys: keys.remove(self._par_name)
            if self._seq_name in keys: keys.remove(self._seq_name)
            if self._cnt_name in keys: keys.remove(self._cnt_name)
        else:
            # Check if datasets in keys are available in the file
            for k in keys:
                if k not in self._fd[group].keys():
                   raise Exception(f'Dataset "/{group}/{k}" does not exist')

        # if group does not already exist, just add it
        if not self._groups or group not in self._groups[:][0]:
            self.check_shape0(group, keys)
            self._groups.append([ group, keys ])
            return

        # if group is already present, need to figure out whether any extra keys need to be added
        for g, k in self._groups:
            if g == group:
                self.check_shape0(group, keys)
                for key in keys:
                    if key not in k:
                        k.append(key)
                return
        raise Exception(f'group "{group}" not found.')

    def keys(self):
        return self._fd.keys()

    def _cols(self,
              group: str,
              key: str) -> List[str]:
        if key == self._par_name: return [ "run", "subrun", "event" ]
        if group in self._colmap and key in self._colmap[group].keys(): return self._colmap[group][key]
        elif self._fd[group][key].shape[1]==1: return [key]
        else: return [ key+"_"+str(c) for c in range(0,self._fd[group][key].shape[1])]

    def get_dataframe(self,
                      group: str,
                      keys: List[str] = []) -> pd.DataFrame:
        if not keys:
            keys = list(self._fd[group].keys())
            if self._seq_name in keys: keys.remove(self._seq_name)
            if self._cnt_name in keys: keys.remove(self._cnt_name)
        dfs = [ pd.DataFrame(np.array(self._fd[group][key]), columns=self._cols(group, key)) for key in keys ]
        return pd.concat(dfs, axis="columns").set_index(["run","subrun","event"])

    def get_dataframe_evt(self,
                          group: str,
                          keys: List[str] = []) -> pd.DataFrame:
        if not keys:
            keys = list(self._data[group].keys())
            if self._seq_name in keys: keys.remove(self._seq_name)
            if self._cnt_name in keys: keys.remove(self._cnt_name)
        dfs = [ pd.DataFrame(np.array(self._data[group][key]), columns=self._cols(group, key)) for key in keys ]
        df = pd.concat(dfs, axis="columns")
        evt_idx_col = []
        for seq in self._seq_cnt[group]:
            evt_idx_col += seq[1]*[seq[0]]
        df['evt_idx'] = evt_idx_col
        return df

    def index(self, idx: int):
        """get the index for a given row"""
        return self._my_index[idx - self._my_start]

    def read_seq(self) -> None:
        for group, datasets in self._groups:
            try:
                # read an HDF5 dataset into a numpy array
                self._whole_seq[group] = np.array(self._fd[group+"/"+self._seq_name])
            except KeyError:
                print(f"Error: dataset {group}/{self._seq_name} does not exist")
                sys.stdout.flush()
                sys.exit(1)

    def read_seq_cnt(self) -> None:
        for group, datasets in self._groups:
            try:
                # read an HDF5 dataset into a numpy array
                self._whole_seq_cnt[group] = np.array(self._fd[group+"/"+self._cnt_name])
            except KeyError:
                print(f"Error: dataset {group}/{self._cnt_name} does not exist")
                sys.stdout.flush()
                sys.exit(1)

    def data_partition(self) -> None:
        # Calculate the start indices and counts of evt.seq assigned to each process
        # self._starts: a numpy array of size nprocs
        # self._counts: a numpy array of size nprocs
        # Note self._starts and self._counts are matter only in root process.
        # self._my_start: (== self._starts[rank]) this process's start
        # self._my_count: (== self._counts[rank]) this process's count
        # self._my_index: partitioned dataset i.e. assigned to this process

        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        nprocs = comm.Get_size()
        self._starts = np.zeros(nprocs, dtype=int)
        self._counts = np.zeros(nprocs, dtype=int)

        if rank == 0:
            if self._use_seq_cnt:
                self.read_seq_cnt()
            else:
                self.read_seq()

            num_events = self._num_events

            if self._evt_part == 0:
                # Below implements event ID based partitioning, which
                # calculates the start and count of evt.seq id for each process
                _count = num_events // nprocs
                for j in range(num_events % nprocs):
                    self._starts[j] = _count * j + j
                    self._counts[j] = _count + 1

                for j in range(num_events % nprocs, nprocs):
                    self._starts[j] = _count * j + num_events % nprocs
                    self._counts[j] = _count

            elif self._evt_part == 1:
                # event amount based partitioning, which calculates event sizes
                # across all groups. Note it is possible multiple consecutive rows
                # a dataset have the same event ID. It is also possible some event
                # IDs contain no data. First, we accumulate numbers of events
                # across all groups
                evt_size = np.zeros(num_events, dtype=int)
                if self._use_seq_cnt:
                    for group, datasets in self._groups:
                        seq_cnt = self._whole_seq_cnt[group]
                        num_datasets = len(datasets)
                        for i in range(seq_cnt.shape[0]):
                            evt_size[seq_cnt[i, 0]] += seq_cnt[i, 1] * num_datasets
                else:
                    for group, datasets in self._groups:
                        seq = self._whole_seq[group]
                        for i in range(seq.shape[0]):
                            evt_size[seq[i, 0]] += 1

                # now we have collected the number of events per event ID across all groups
                total_evt_num = np.sum(evt_size)
                avg_evt_num = total_evt_num // nprocs
                avg_evt = total_evt_num // num_events / 2

                # assign ranges of event IDs to individual processes
                acc_evt_num = 0
                rank_id = 0
                for j in range(num_events):
                    if rank_id == nprocs - 1: break
                    if acc_evt_num + evt_size[j] >= avg_evt_num:
                        remain_l = avg_evt_num - acc_evt_num
                        remain_r = evt_size[j] - remain_l
                        if remain_l > remain_r and remain_l > avg_evt:
                            # assign event j to rank_id
                            self._counts[rank_id] += 1
                            acc_evt_num = 0
                        else:
                            # assign event j to rank_id+1
                            self._counts[rank_id+1] = 1
                            acc_evt_num = evt_size[j]
                        # done with rank_id i
                        rank_id += 1
                        self._starts[rank_id] = self._starts[rank_id-1] + self._counts[rank_id-1]
                    else:
                        self._counts[rank_id] += 1
                        acc_evt_num += evt_size[j]
                self._counts[nprocs-1] += num_events - j

            elif self._evt_part == 2:
                # use event amounts in the particle_table only to partition events
                seq_cnt = self._whole_seq_cnt['particle_table']
                total_evt_num = np.sum(seq_cnt[:,1])
                avg_evt_num = total_evt_num // nprocs
                avg_evt = total_evt_num // seq_cnt.shape[0] / 2

                self._starts[0] = seq_cnt[0,0]
                acc_evt_num = 0
                rank_id = 0
                for j in range(seq_cnt.shape[0]):
                    if rank_id == nprocs - 1: break
                    if acc_evt_num + seq_cnt[j,1] >= avg_evt_num:
                        remain_l = avg_evt_num - acc_evt_num
                        remain_r = seq_cnt[j,1] - remain_l
                        # if remain_r > remain_l:
                        if remain_l > remain_r and remain_l > avg_evt:
                            # assign event j to rank_id
                            self._counts[rank_id] = seq_cnt[j+1, 0] - self._starts[rank_id]
                            self._starts[rank_id+1] = seq_cnt[j+1, 0]
                            acc_evt_num = 0
                        else:
                            # assign event j to rank_id+1
                            self._counts[rank_id] = seq_cnt[j, 0] - self._starts[rank_id]
                            self._starts[rank_id+1] = seq_cnt[j, 0]
                            acc_evt_num = seq_cnt[j, 1]
                        # done with rank_id
                        rank_id += 1
                    else:
                        acc_evt_num += seq_cnt[j, 1]

                self._counts[nprocs-1] = num_events - self._starts[nprocs-1]

        # All processes participate the collective communication, scatter.
        # Root distributes start and count to all processes. Note only root process
        # uses self._starts and self._counts.
        start_count = np.empty([nprocs, 2], dtype=int)
        start_count[:, 0] = self._starts[:]
        start_count[:, 1] = self._counts[:]
        recvbuf = np.empty(2, dtype=int)
        comm.Scatter(start_count, recvbuf, root=0)
        self._my_start = recvbuf[0]
        self._my_count = recvbuf[1]

        # This process is assigned event IDs of range from self._my_start to
        # (self._my_start + self._my_count - 1)
        # print("my_start=",self._my_start," my_count=",self._my_count);

        # each process reads its share of dataset and stores it in a numpy
        # array
        self._my_index = np.array(self._index[self._my_start : self._my_start + self._my_count, :])

    def binary_search_min(self, key, base, nmemb):
        low = 0
        high = nmemb
        while low != high:
                mid = (low + high) // 2
                if base[mid] < key:
                        low = mid + 1
                else:
                        high = mid
        return low

    def binary_search_max(self, key, base, nmemb):
        low = 0
        high = nmemb
        while low != high:
                mid = (low + high) // 2
                if base[mid] <= key:
                        low = mid + 1
                else:
                        high = mid
        return (low - 1)

    def calc_bound_seq(self, group):
        # return the lower and upper array indices of subarray assigned to this
        # process, using the partition sequence dataset

        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        nprocs = comm.Get_size()

        displ  = np.zeros([nprocs], dtype=int)
        count  = np.zeros([nprocs], dtype=int)
        bounds = np.zeros([nprocs, 2], dtype=int)

        all_evt_seq = None
        if rank == 0:
            # root reads the entire dataset self._seq_name, if not already
            if not self._whole_seq: self.read_seq()

            all_evt_seq = self._whole_seq[group]
            dim = len(all_evt_seq)

            # calculate displ, count to be used in scatterV for all processes
            for i in range(nprocs):
                if self._counts[i] == 0: continue
                end = self._starts[i] + self._counts[i] - 1
                bounds[i, 0] = self.binary_search_min(self._starts[i], all_evt_seq, dim)
                bounds[i, 1] = self.binary_search_max(end,             all_evt_seq, dim)
                displ[i] = bounds[i, 0]
                count[i] = bounds[i, 1] - bounds[i, 0] + 1

        lower_upper = np.empty([2], dtype=int)

        # root distributes start and end indices to all processes
        comm.Scatter(bounds, lower_upper, root=0)

        # this process is assigned array indices from lower to upper
        # print("group=",group," lower=",lower," upper=",upper," count=",upper-lower)

        lower = 0
        upper = 0
        if self._my_count > 0:
            lower = lower_upper[0]
            upper = lower_upper[1] + 1

        # root scatters the subarray of evt_seq to all processes
        self._evt_seq[group] = np.zeros(upper - lower, dtype=np.int64)
        comm.Scatterv([all_evt_seq, count, displ, MPI.LONG_LONG], self._evt_seq[group], root=0)

        return lower, upper

    def calc_bound_seq_cnt(self, group):
        # return the lower and upper array indices of subarray assigned to this
        # process, using the partition sequence-count dataset

        comm   = MPI.COMM_WORLD
        rank   = comm.Get_rank()
        nprocs = comm.Get_size()

        displ   = np.zeros([nprocs], dtype=int)
        count   = np.zeros([nprocs], dtype=int)
        seq_cnt = np.zeros([nprocs, 2], dtype=int)

        all_seq_cnt = None
        if rank == 0:
            # root reads the entire dataset self._cnt_name, if not already
            if not self._whole_seq_cnt: self.read_seq_cnt()

            all_seq_cnt = self._whole_seq_cnt[group]
            dim = len(all_seq_cnt)

            # calculate displ, count for all processes to be used in scatterV
            recv_rank = 0  # receiver rank
            displ[recv_rank] = 0
            seq_cnt[recv_rank, 0] = 0
            seq_end = self._starts[recv_rank] + self._counts[recv_rank]
            seq_id = 0
            for i in range(dim):
                if all_seq_cnt[i, 0] >= seq_end :
                    seq_cnt[recv_rank, 1] = i - displ[recv_rank]
                    recv_rank += 1  # move on to the next receiver rank
                    seq_end = self._starts[recv_rank] + self._counts[recv_rank]
                    displ[recv_rank] = i
                    seq_cnt[recv_rank, 0] = seq_id
                seq_id += all_seq_cnt[i, 1]

            # last receiver rank
            seq_cnt[recv_rank, 1] = dim - displ[recv_rank]

            # print("starts=",self._starts," counts=",self._counts," displ=",displ," count=",count," seq_cnt=",seq_cnt )
            displ[:] *= 2
            count[:] = seq_cnt[:, 1] * 2

        # root distributes seq_cnt to all processes
        my_seq_cnt = np.empty([2], dtype=int)
        comm.Scatter(seq_cnt, my_seq_cnt, root=0)

        # self._seq_cnt[group][:, 0] is the event ID
        # self._seq_cnt[group][:, 1] is the number of elements
        self._seq_cnt[group] = np.empty([my_seq_cnt[1], 2], dtype=np.int64)

        # root scatters the subarray of evt_seq to all processes
        comm.Scatterv([all_seq_cnt, count, displ, MPI.LONG_LONG], self._seq_cnt[group], root=0)

        lower = 0
        upper = 0
        if self._my_count > 0:
            lower = my_seq_cnt[0]
            upper = my_seq_cnt[0] + np.sum(self._seq_cnt[group][:, 1])

        # this process is assigned array indices from lower to upper
        # print("group=",group," lower=",lower," upper=",upper," count=",upper-lower)

        return lower, upper

    def read_data(self,
                  start: int,
                  count: int) -> None:
        # (sequentially) read subarrays of all datasets in all groups that fall
        # in the range of self._seq_name, starting from 'start' and amount of 'count'

        for group, datasets in self._groups:
            if self._use_seq_cnt:
                # use evt_id.seq_cnt to calculate subarray boundaries
                # reads the entire dataset self._cnt_name, if not already
                if not self._whole_seq_cnt or group not in self._whole_seq_cnt.keys():
                    self.read_seq_cnt()
                all_seq_cnt = self._whole_seq_cnt[group]
                # search indices of start and end in all_seq_cnt
                # all_seq_cnt[:,0] are all unique
                ilower = np.searchsorted(all_seq_cnt[:,0], start)
                iupper = np.searchsorted(all_seq_cnt[:,0], start+count)
                self._seq_cnt[group] = np.array(all_seq_cnt[ilower:iupper], dtype=np.int64)
                lower = np.sum(all_seq_cnt[0:ilower, 1])
                upper = lower + np.sum(all_seq_cnt[ilower:iupper, 1])
            else:
                # use evt_id.seq to calculate subarray boundaries
                # root reads the entire dataset self._seq_name, if not already
                if not self._whole_seq: self.read_seq()
                all_evt_seq = self._whole_seq[group]
                dim = len(all_evt_seq)
                # search indices of start and end in all_seq
                # all_seq[:] are not unique
                end = start + count - 1
                lower = self.binary_search_min(start, all_evt_seq, dim)
                upper = self.binary_search_max(end,   all_evt_seq, dim)
                upper += 1
                self._evt_seq[group] = np.array(all_evt_seq[lower:upper], dtype=np.int64)

            # print("read_data - group=",group,", lower=",lower," upper=",upper)

            # Iterate through all the datasets and read the subarray from index lower
            # to upper and store it into a dictionary with the names of group and
            # dataset as the key.
            self._data[group] = {}
            for dset in datasets:
                # read subarray into a numpy array
                self._data[group][dset] = np.array(self._fd[group][dset][lower : upper])

        self._my_start = start
        self._my_count = count
        # read assigned partitioning key dataset into a numpy array
        self._my_index = np.array(self._index[start : start + count, :])

    def read_data_all(self,
                      use_seq_cnt: bool = True,
                      evt_part: int = 2,
                      profile: bool = False) -> None:
        # use_seq_cnt: True  - use event.seq_cnt dataset to calculate partitioning
        #                      starts and counts
        #              False - use event.seq dataset to calculate starts and counts
        # evt_part: 0  - partition based on event IDs
        #           1 - partition based on event amount
        #           2 - partition based on event amount of particle table (default)
        # Parallel read dataset subarrays assigned to this process ranging from
        # array index of self._my_start to (self._my_start + self._my_count - 1)
        if profile:
            par_time = 0
            bnd_time = 0
            rds_time = 0
            time_s = MPI.Wtime()

        self._use_seq_cnt = use_seq_cnt
        self._evt_part = evt_part

        # calculate the data partitioning start indices and amounts assigned to
        # each process. Set self._starts, self._counts, self._my_start,
        # self._my_count, and self._my_index
        self.data_partition()

        if profile:
            time_e = MPI.Wtime()
            par_time = time_e - time_s
            time_s = time_e

        for group, datasets in self._groups:
            if self._use_seq_cnt:
                # use evt_id.seq_cnt to calculate subarray boundaries
                lower, upper = self.calc_bound_seq_cnt(group)
            else:
                # use evt_id.seq to calculate subarray boundaries
                lower, upper = self.calc_bound_seq(group)

            if profile:
                time_e = MPI.Wtime()
                bnd_time += time_e - time_s
                time_s = time_e

            # print("read_data_all - group=",group,", lower=",lower," upper=",upper)

            # Iterate through all the datasets and read the subarray from index lower
            # to upper and store it into a dictionary with the names of group and
            # dataset as the key.
            self._data[group] = {}
            for dset in datasets:
                # read subarray into a numpy array
                self._data[group][dset] = np.array(self._fd[group][dset][lower : upper])

            if profile:
                time_e = MPI.Wtime()
                rds_time += time_e - time_s
                time_s = time_e

        if profile:
            rank   = MPI.COMM_WORLD.Get_rank()
            nprocs = MPI.COMM_WORLD.Get_size()

            total_t = np.array([par_time, bnd_time, rds_time])
            max_total_t = np.zeros(3)
            MPI.COMM_WORLD.Reduce(total_t, max_total_t, op=MPI.MAX, root = 0)
            min_total_t = np.zeros(3)
            MPI.COMM_WORLD.Reduce(total_t, min_total_t, op=MPI.MIN, root = 0)
            if rank == 0:
                print("---- Timing break down of the file read phase (in seconds) -------")
                if self._use_seq_cnt:
                    print(f'Use "{self._cnt_name}" to calculate subarray boundaries')
                else:
                    print(f'Use "{self._seq_name}" to calculate subarray boundaries')

                print("data partitioning           time ", end='')
                print("MAX=%8.2f  MIN=%8.2f" % (max_total_t[0], min_total_t[0]))
                print("calc boundaries             time ", end='')
                print("MAX=%8.2f  MIN=%8.2f" % (max_total_t[1], min_total_t[1]))
                print("read datasets               time ", end='')
                print("MAX=%8.2f  MIN=%8.2f" % (max_total_t[2], min_total_t[2]))
                print("(MAX and MIN timings are among %d processes)" % nprocs)

    def build_evt(self,
                  start: int = None,
                  count: int = None) -> List[Dict]:
        # This process is responsible for event IDs from start to (start+count-1).
        # All data of the same event ID will be used to create a graph.
        # This function collects all data based on self._seq_name, or
        # self._cnt_name into a python list containing Pandas DataFrames, one
        # for a unique event ID.
        if not self._groups:
            raise Exception('cannot build event without adding any HDF5 groups')

        ret_list = []

        if start is None: start = self._my_start
        if count is None: count = self._my_count

        num_miss = 0

        # track the latest used index per group
        idx_grp = dict.fromkeys(self._data.keys(), 0)

        # accumulate starting array index per group
        idx_start = dict.fromkeys(self._data.keys(), 0)

        # Iterate through assigned event IDs
        for idx in range(int(start), int(start+count)):
            # check if idx is missing in all groups
            is_missing = True
            if self._use_seq_cnt:
                for group in self._data.keys():
                    if idx_grp[group] >= len(self._seq_cnt[group][:, 0]):
                        continue
                    if idx == self._seq_cnt[group][idx_grp[group], 0]:
                        is_missing = False
                        break
            else:
                for group in self._data.keys():
                    dim = len(self._evt_seq[group])
                    lower = self.binary_search_min(idx, self._evt_seq[group], dim)
                    upper = self.binary_search_max(idx, self._evt_seq[group], dim) + 1
                    if lower < upper:
                        is_missing = False
                        break

            # this idx is missing in all groups
            if is_missing:
                num_miss += 1
                continue

            # for each event seq ID, create a dictionary, ret
            #   first item: key is "index" and value is the event seq ID
            #   remaining items: key is group name and value is a Pandas DataFrame
            #   containing the dataset subarray in this group with the event ID, idx
            ret = Event(idx, self.index(idx))

            # Iterate through all groups
            for group in self._data.keys():

                if self._use_seq_cnt:
                    # self._seq_cnt[group][:, 0] is the event ID
                    # self._seq_cnt[group][:, 1] is the number of elements

                    if idx_grp[group] >= len(self._seq_cnt[group][:, 0]) or idx < self._seq_cnt[group][idx_grp[group], 0]:
                        # idx is missing from this group but may not in other groups
                        # create an empty Pandas DataFrame
                        dfs = []
                        for dataset in self._data[group].keys():
                            data_dataframe = pd.DataFrame(columns=self._cols(group, dataset))
                            dfs.append(data_dataframe)
                        ret[group] = pd.concat(dfs, axis="columns")
                        # print("xxx",group," missing idx=",idx," _seq_cnt[0]=",self._seq_cnt[group][idx_grp[group], 0])
                        continue

                    lower = idx_start[group]
                    upper = self._seq_cnt[group][idx_grp[group], 1] + lower

                    # print("group=",group," idx=",idx," lower=",lower," upper=",upper)
                    idx_start[group] += self._seq_cnt[group][idx_grp[group], 1]
                    idx_grp[group] += 1

                else:
                    # Note self._evt_seq stores event ID values and is already sorted in
                    # an increasing order
                    dim = len(self._evt_seq[group])

                    # Find the local start and end row indices for this event ID, idx
                    lower = self.binary_search_min(idx, self._evt_seq[group], dim)
                    upper = self.binary_search_max(idx, self._evt_seq[group], dim) + 1
                    # print("For idx=",idx, " lower=",lower, " upper=",upper)

                # dfs is a python list containing Pandas DataFrame objects
                dfs = []
                for dataset in self._data[group].keys():
                    if lower >= upper:
                        # idx is missing from the dataset self._seq_name,
                        # In this case, create an empty numpy array
                        data = np.array([])
                    else:
                        # array elements from lower to upper of this dataset have the
                        # event ID == idx
                        data = self._data[group][dataset][lower : upper]

                    # create a Pandas DataFrame to store the numpy array
                    df = pd.DataFrame(data, columns=self._cols(group, dataset))
                    for col in df.columns:
                        if df[col].dtype == '|S64' or df[col].dtype == 'object':
                            df[col] = df[col].str.decode('utf-8')
                    dfs.append(df)

                # concate into the dictionary "ret" with group names as keys
                ret[group] = pd.concat(dfs, axis="columns")

            # Add all dictionaries "ret" into a list.
            # Each of them corresponds to the data of one single event ID
            ret_list.append(ret)

        # print("start=",start," count=",count," num miss IDs=",num_miss)
        return ret_list

    def process(self,
                processor: Callable[[Event], Tuple[str, Any]],
                out: Callable[[Any, str], None]) -> None:
        '''Process all events in this data partition'''
        comm = MPI.COMM_WORLD
        nprocs = comm.Get_size()
        rank = comm.Get_rank()
        if rank == 0:
            out.write_metadata(processor.metadata)
        self.read_data_all()
        evt_list = self.build_evt()
        for evt in evt_list:
            name, data = processor(evt)
            if data is not None: out(name, data)


import pandas as pd
import particle

class StandardLabels:

    def __init__(self,
                 gamma_threshold: float = 0.02,
                 hadron_threshold: float = 0.2):
        self._labels = [
            'pion',
            'muon',
            'kaon',
            'hadron',
            'shower',
            'michel',
            'diffuse',
            'invisible'
        ]
        self._gamma_threshold = gamma_threshold
        self._hadron_threshold = hadron_threshold

    @property
    def labels(self):
        return self._labels

    def label(self, idx: int):
        if not 0 <= label < len(self._labels):
            raise Exception(f'index {idx} out of range for {len(self._labels)} labels.')
        return self._labels[idx]

    def index(self, name: str):
        if name not in self._labels:
            raise Exception(f'"{name}" is not the name of a class.')
        return self._labels.index(name)
    
    @property
    def pion(self):
        return self.index('pion')

    @property
    def muon(self):
        return self.index('muon')

    @property
    def kaon(self):
        return self.index('kaon')

    @property
    def hadron(self):
        return self.index('hadron')

    @property
    def shower(self):
        return self.index('shower')

    @property
    def michel(self):
        return self.index('michel')

    @property
    def diffuse(self):
        return self.index('diffuse')
    
    @property
    def invisible(self):
        return self.index('invisible')

    def __call__(self,
                 part: pd.DataFrame):
        '''Standard labelling function.

        Pion, Muon, Kaon, Hadron, EM shower, Michel electron,
        diffuse activity.
        '''

        def walk(part, particles, depth, sl, il):
            def s(part, particles):
                sl, slc = -1, None
                parent_type = 0 if part.parent_id == 0 else particles.type[part.parent_id]

                def pion_labeler(part, parent_type):
                    sl = self.pion
                    slc = None
                    return sl, slc

                def muon_labeler(part, parent_type):
                    sl = self.muon
                    slc = None
                    return sl, slc

                def kaon_labeler(part, parent_type):
                    sl = self.kaon
                    slc = None
                    return sl, slc

                def neutral_pions_kaons_labeler(part, parent_type):
                    sl = self.invisible
                    slc = None
                    return sl, slc

                def electron_positron_labeler(part, parent_type):
                    if part.start_process == 'primary':
                        sl = self.shower
                        slc = self.shower
                    elif abs(parent_type) == 13 and (part.start_process == 'muMinusCaptureAtRest' \
                        or part.start_process == 'muPlusCaptureAtRest' or part.start_process == 'Decay'):
                        sl = self.michel
                        slc = self.michel
                    elif part.start_process == 'conv' or part.end_process == 'conv' \
                        or part.start_process == 'compt' or part.end_process == 'compt':
                        if part.momentum >= self._gamma_threshold:
                            sl = self.shower
                            slc = self.shower
                        else:
                            sl = self.diffuse
                            slc = self.diffuse
                    elif part.start_process == 'eBrem' or part.end_process == 'phot' \
                        or part.end_process == 'photonNuclear' or part.end_process == 'eIoni':
                        sl = self.diffuse
                        slc = None
                    elif part.start_process == 'muIoni' or part.start_process == 'hIoni' \
                        or part.start_process == 'eIoni':
                        if part.start_process == 'muIoni':
                            sl = self.muon
                            slc = None
                        elif part.start_process == 'hIoni':
                            if abs(parent_type) == 2212:
                                sl = self.hadron
                                if part.momentum <= 0.0015: sl = self.diffuse
                            else:
                                sl = self.pion
                            slc = None
                        else:
                            sl = self.diffuse
                            slc = None
                    elif part.end_process == 'StepLimiter' or part.end_process == 'annihil' \
                        or part.end_process == 'eBrem' or part.start_process == 'hBertiniCaptureAtRest' \
                        or part.end_process == 'FastScintillation' or part.start_process == 'muPairProd' \
                        or part.start_process == 'phot':
                        sl = self.diffuse
                        slc = self.diffuse
                    else:
                        raise Exception(f'labelling failed for electron with start process "{part.start_process}" and end process "{part.end_process}')

                    return sl, slc

                def gamma_labeler(part, parent_type):
                    if part.start_process == 'conv' or part.end_process == 'conv' \
                        or part.start_process == 'compt' or part.end_process == 'compt':
                        if part.momentum >= self._gamma_threshold:
                            sl = self.shower
                            slc = self.shower
                        else:
                            sl = self.diffuse
                            slc = self.diffuse
                    elif part.start_process == 'eBrem' or part.end_process == 'phot' \
                        or part.end_process == 'photonNuclear':
                        sl = self.diffuse
                        slc = None
                    else:
                        raise Exception(f'labelling failed for photon with start process "{part.start_process}" and end process "{part.end_process}')
                    return sl, slc

                def unlabeled_particle(part, parent_type):
                    raise Exception(f"particle not recognised! PDG code {part.type}, parent PDG code {parent_type}, start process {part.start_process}, end process {part.end_process}")

                particle_processor = {
                    211: pion_labeler,
                    221: pion_labeler,
                    331: pion_labeler,
                    223: pion_labeler,
                    13: muon_labeler,
                    321: kaon_labeler,
                    111: neutral_pions_kaons_labeler,
                    311: neutral_pions_kaons_labeler,
                    310: neutral_pions_kaons_labeler,
                    130: neutral_pions_kaons_labeler,
                    113: neutral_pions_kaons_labeler,
                    411: kaon_labeler, # D meson
                    11: electron_positron_labeler,
                    22: gamma_labeler
                }

                if particle.pdgid.charge(part.type) == 0 and part.end_process == 'CoupledTransportation':
                    # neutral particle left the volume boundary
                    sl = self.invisible
                else:
                    func = particle_processor.get(abs(part.type), lambda x ,y: (-1, None))
                    sl, slc = func(part, parent_type)

                # baryon interactions - hadron or diffuse
                if (particle.pdgid.is_baryon(part.type) and particle.pdgid.charge(part.type) == 0) \
                    or particle.pdgid.is_nucleus(part.type):
                    sl = self.diffuse
                if particle.pdgid.is_baryon(part.type) and particle.pdgid.charge(part.type) != 0:
                    if abs(part.type) == 2212 and part.momentum >= self._hadron_threshold:
                        sl = self.hadron
                    else:
                        sl = self.diffuse

                # call a charged tau highly ionising - should revisit this
                if abs(part.type) == 15:
                    sl = self.hadron

                # check to make sure particle was assigned
                if sl == -1:
                    unlabeled_particle(part, parent_type)

                return sl, slc

            def i(part, particles, sl):
                il, ilc = -1, None
                if sl == self.muon and part.start_process == 'muIoni':
                    il = part.parent_id
                elif (sl == self.pion or sl == self.hadron) and part.start_process == 'hIoni':
                    il = part.parent_id
                elif sl != self.diffuse and sl != self.invisible:
                    il = part.g4_id
                    if sl == self.shower: ilc = il
                    if sl == self.michel: ilc = il
                return il, ilc

            if sl is not None: slc = sl
            else: sl, slc = s(part, particles)

            if il is not None: ilc = il
            else: il, ilc = i(part, particles, sl)

            ret = [ {
                "g4_id": part.g4_id,
                "parent_id": part.parent_id,
                "type": part.type,
                "start_process": part.start_process,
                "end_process": part.end_process,
                "momentum": part.momentum,
                "semantic_label": sl,
                "instance_label": il } ]
            for _, row in particles[(part.g4_id==particles.parent_id)].iterrows():
                ret += walk(row, particles, depth+1, slc, ilc)
            return ret

        ret = []
        part = part.set_index("g4_id", drop=False)
        primaries = part[(part.parent_id==0)]
        for _, primary in primaries.iterrows():
            ret += walk(primary, part, 0, None, None)
        if len(ret)==0: return
        labels = pd.DataFrame.from_dict(ret)
        instances = { val: i for i, val in enumerate(labels[(labels.instance_label>=0)].instance_label.unique()) }

        def alias_instance(row, instances):
            if row.instance_label == -1: return -1
            return instances[row.instance_label]

        labels["instance_label"] = labels.apply(alias_instance, args=[instances], axis="columns")
        return labels

    def validate(self, labels: pd.Series):
        mask = (labels < 0) | (labels >= len(self._labels) - 1)
        if mask.any():
            raise Exception(f'{mask.sum()} semantic labels are out of range: {labels[mask]}.')



import pynuml

def process_event(key, out, sp, hit, part, edep, l=standard, voxelsize=1):
    """Process an event into a 3D pixel map"""
    import numpy as np, torch, MinkowskiEngine as ME

    # skip any events with no simulated hits
    if (hit.index==key).sum() == 0: return
    if (edep.index==key).sum() == 0: return

    # label true particles
    evt_part = part.loc[key].reset_index(drop=True)
    evt_part = l.panoptic_label(evt_part)

    # get energy depositions and ground truth
    evt_edep = edep.loc[key].reset_index(drop=True)
    evt_edep = evt_edep.merge(evt_part[["g4_id", "semantic_label"]], on="g4_id", how="left").drop("g4_id", axis="columns")
    scores = evt_edep.groupby(["hit_id", "semantic_label"]).agg({"energy": "sum"}).reset_index()

    # class number and names
    n = len(l.label) - 1
    lnames = [ it.name for it in l.label ][:-1]
    noise = np.zeros(n)
    noise[l.label.diffuse.value] = 1

    def fractional_truth(row, n):
        label = np.zeros(n)
        label[int(row.semantic_label)] = row.energy
        return label
    scores["slabel"] = scores.apply(fractional_truth, args=[n], axis="columns")
    scores = scores.groupby("hit_id").agg({"slabel": "sum"})

    # Propagate labels to hits
    evt_hit = hit.loc[key].reset_index(drop=True).merge(scores, on="hit_id", how="inner")
    evt_sp = sp.loc[key].reset_index(drop=True)

    # skip events with fewer than 50 simulated hits in any plane, or fewer than 50 spacepoints
    for i in range(3):
        if (evt_hit.global_plane==i).sum() < 50: return
    if evt_sp.shape[0] < 50: return

    # merge hits into spacepoints
    for plane in ["u","v","y"]:
        evt_sp = evt_sp.merge(evt_hit[["hit_id","integral","slabel"]].add_suffix(f"_{plane}"), on=f"hit_id_{plane}", how="left")
        evt_sp[f"integral_{plane}"] = evt_sp[f"integral_{plane}"].fillna(0)

    def merge_truth(row, n):
        labels = np.zeros(n)
        for plane in ["u","v","y"]:
            vals = row[f"slabel_{plane}"]
            if type(vals) != float: labels += vals
        return labels

    evt_sp["slabel"] = evt_sp.apply(merge_truth, args=[len(l.label)-1], axis="columns")
    evt_sp = evt_sp[["slabel", "position_x", "position_y", "position_z", "integral_u", "integral_v", "integral_y"]]

    # voxelise spacepoints and aggregate labels
    def voxelise(row):
        return np.floor(row.position_x/voxelsize), np.floor(row.position_y/voxelsize), np.floor(row.position_z/voxelsize)
    evt_sp["c"] = evt_sp.apply(voxelise, axis="columns")
    evt_sp = evt_sp.drop(["position_x", "position_y", "position_z"], axis="columns")
    evt_sp = evt_sp.groupby("c").agg({"integral_u": "sum", "integral_v": "sum", "integral_y": "sum", "slabel": "sum"}).reset_index()
    def norm_truth(row, noise):
        lsum = row.slabel.sum()
        return noise if lsum == 0 else row.slabel / lsum
    evt_sp["slabel"] = evt_sp.apply(norm_truth, args=[noise], axis="columns")

    spm = {
        "f": torch.tensor(evt_sp[["integral_u", "integral_v", "integral_y"]].to_numpy()).float(),
        "c": torch.tensor(evt_sp["c"]).int(),
        "ys": torch.tensor(evt_sp["slabel"]).float()
    }
    out.save(spm, f"r{key[0]}_sr{key[1]}_evt{key[2]}")

def process_file(out, fname, p=process_event, l=standard, voxelsize=1):
    """Process all events in a file into graphs"""
    f = NuMLFile(fname)

    evt = f.get_dataframe("event_table", ["event_id"])
    sp = f.get_dataframe("spacepoint_table")
    hit = f.get_dataframe("hit_table")
    part = f.get_dataframe("particle_table", ["event_id", "g4_id", "parent_id", "type", "momentum", "start_process", "end_process"])
    edep = f.get_dataframe("edep_table")

    # loop over events in file
    for key in evt.index: p(key, out, sp, hit, part, edep, l, voxelsize)




import pandas as pd

class PDKLabels:
    def __init__(self):
        self._labels = ('nu', 'pdk')

    @property
    def labels(self):
        return self._labels

    def label(self, idx: int):
        if not 0 <= label < len(self._labels):
            raise Exception(f'index {idx} out of range for {len(self._labels)} labels.')
        return self._labels[idx]

    def index(self, name: str):
        if name not in self._labels:
            raise Exception(f'"{name}" is not the name of a class.')
        return self._labels.index(name)

    @property
    def nu(self):
        return self.index('nu')

    @property
    def pdk(self):
        return self.index('pdk')

    def __call__(self, event: pd.Series):
        if 12 <= abs(event.nu_pdg) <= 16:
            return self.nu
        else:
            return self.pdk


from abc import ABC
from typing import Any, Dict, List, Tuple

from ..io import File

class ProcessorBase(ABC):
    '''Base class for event processing'''

    def __init__(self, file: File):
        for group, keys in self.columns.items():
            file.add_group(group, keys)

    @property
    def columns(self) -> Dict[str, List[str]]:
        raise NotImplementedError

    def __call__(self, evt: Any) -> Tuple[str, Any]:
        raise NotImplementedError


import pandas as pd
from torch_geometric.data import Batch, HeteroData
import plotly.express as px
from plotly.graph_objects import FigureWidget
import warnings

class GraphPlot:
    def __init__(self,
                 planes: list[str],
                 classes: list[str],
                 filter_threshold: float = 0.5):
        self._planes = planes
        self._classes = classes
        self._labels = pd.CategoricalDtype(['background']+classes, ordered=True)
        self._cmap = { c: px.colors.qualitative.Plotly[i] for i, c in enumerate(classes) }
        self._cmap['background'] = 'lightgrey'
        self.filter_threshold = filter_threshold

        # temporarily silence this pandas warning triggered by plotly,
        # which we don't have any power to fix but will presumably
        # be fixed on their end at some point
        warnings.filterwarnings("ignore", ".*The default of observed=False is deprecated and will be changed to True in a future version of pandas.*")
        self._truth_cols = ( 'g4_id', 'parent_id', 'pdg' )

    def to_dataframe(self, data: HeteroData):
        def to_categorical(arr):
            return pd.Categorical.from_codes(codes=arr+1, dtype=self._labels)
        if isinstance(data, Batch):
            raise Exception('to_dataframe does not support batches!')
        dfs = []
        for p in self._planes:
            plane = data[p].to_dict()
            df = pd.DataFrame(plane['id'], columns=['id'])
            df['plane'] = p
            df[['wire','time']] = plane['pos']
            if "c" in plane:
                df[["x", "y", "z"]] = plane["c"]
            df['y_filter'] = plane['y_semantic'] != -1
            mask = df.y_filter.values
            df['y_semantic'] = to_categorical(plane['y_semantic'])
            df['y_instance'] = plane['y_instance'].numpy().astype(str)

            # add detailed truth information if it's available
            for col in self._truth_cols:
                if col in plane.keys():
                    df[col] = plane[col].numpy()

            # add model prediction if it's available
            if 'x_semantic' in plane.keys():
                df['x_semantic'] = to_categorical(plane['x_semantic'].argmax(dim=-1).detach())
                df[self._classes] = plane['x_semantic'].detach()
            if 'x_filter' in plane.keys():
                df['x_filter'] = plane['x_filter'].detach()

            dfs.append(df)
        df = pd.concat(dfs)
        md = data['metadata']
        df['run'] = md.run.item()
        df['subrun'] = md.subrun.item()
        df['event'] = md.event.item()
        return df

    def plot(self,
             data: HeteroData,
             target: str = 'hits',
             how: str = 'none',
             filter: str = 'show',
             xyz: bool = False,
             width: int = None,
             height: int = None,
             title: bool = True) -> FigureWidget:

        df = self.to_dataframe(data)

        # no colour
        if target == 'hits':
            opts = {
                'title': 'Graph hits',
            }

        # semantic labels
        elif target == 'semantic':
            if how == 'true':
                opts = {
                    'title': 'True semantic labels',
                    'labels': { 'y_semantic': 'Semantic label' },
                    'color': 'y_semantic',
                    'color_discrete_map': self._cmap,
                }
            elif how == 'pred':
                opts = {
                    'title': 'Predicted semantic labels',
                    'labels': { 'x_semantic': 'Semantic label' },
                    'color': 'x_semantic',
                    'color_discrete_map': self._cmap,
                }
            elif how in self._classes:
                opts = {
                    'title': f'Predicted semantic label strength for {how} class',
                    'labels': { how: f'{how} probability' },
                    'color': how,
                    'color_continuous_scale': px.colors.sequential.Reds,
                }
            else:
                raise Exception('for semantic labels, "how" must be one of "true", "pred" or the name of a class.')

        # instance labels
        elif target == 'instance':
            if how == 'true':
                opts = {
                    'title': 'True instance labels',
                    'labels': { 'y_instance': 'Instance label' },
                    'color': 'y_instance',
                    'symbol': 'y_semantic',
                    'color_discrete_map': self._cmap,
                }
            elif how == 'pred':
                opts = {
                    'title': 'Predicted instance labels',
                    'labels': { 'x_instance': 'Instance label' },
                    'color': 'x_instance',
                    'symbol': 'x_semantic',
                    'color_discrete_map': self._cmap,
                }
            else:
                raise Exception('for instance labels, "how" must be one of "true" or "pred".')

        # filter labels
        elif target == 'filter':
            if how == 'true':
                opts = {
                    'title': 'True filter labels',
                    'labels': { 'y_filter': 'Filter label' },
                    'color': 'y_filter',
                    'color_discrete_map': { 0: 'coral', 1: 'mediumseagreen' },
                }
            elif how == 'pred':
                opts = {
                    'title': 'Predicted filter labels',
                    'labels': { 'x_filter': 'Filter label' },
                    'color': 'x_filter',
                    'color_continuous_scale': px.colors.sequential.Reds,
                }
            else:
                raise Exception('for filter labels, "how" must be one of "true" or "pred".')

        else:
            raise Exception('"target" must be one of "hits", "semantic", "instance" or "filter".')

        if filter == 'none':
            # don't do any filtering
            pass
        elif filter == 'show':
            # show hits predicted to be background in grey
            if target == 'semantic' and how == 'pred':
                df.x_semantic[df.x_filter < self.filter_threshold] = 'background'
        elif filter == 'true':
            # remove true background hits
            df = df[df.y_filter.values]
            opts['title'] += ' (filtered by truth)'
        elif filter == 'pred':
            # remove predicted background hits
            df = df[df.x_filter > self.filter_threshold]
            opts['title'] += ' (filtered by prediction)'
        else:
            raise Exception('"filter" must be one of "none", "show", "true" or "pred".')

        if not title:
            opts.pop('title')

        # set hover data
        opts['hover_data'] = {
            'y_semantic': True,
            'wire': ':.1f',
            'time': ':.1f',
        }
        opts['labels'] = {
            'y_filter': 'filter truth',
            'y_semantic': 'semantic truth',
            'y_instance': 'instance truth',
        }
        if 'x_filter' in df:
            opts['hover_data']['x_filter'] = True
            opts['labels']['x_filter'] = 'filter prediction'
        if 'x_semantic' in df:
            opts['hover_data']['x_semantic'] = True
            opts['labels']['x_semantic'] = 'semantic prediction'
        if 'x_instance' in df:
            opts['hover_data']['x_instance'] = ':.4f'
            opts['labels']['x_instance'] = 'instance prediction'
        for col in self._truth_cols:
            if col in df:
                opts['hover_data'][col] = True

        if xyz:
            fig = px.scatter_3d(df, x="x", y="y", z="z",
                                width=width, height=height, **opts)
            fig.update_traces(marker_size=1)
        else:
            fig = px.scatter(df, x='wire', y='time', facet_col='plane',
                            width=width, height=height, **opts)
            fig.update_xaxes(matches=None)
            for a in fig.layout.annotations:
                a.text = a.text.replace('plane=', '')

        # set the legend to horizontal
        fig.update_layout(
            legend_orientation='h',
            legend_yanchor='bottom', legend_y=1.05,
            legend_xanchor='right', legend_x=1,
            margin_l=20, margin_r=20, margin_t=20, margin_b=20,
            title_automargin=title,
        )

        return FigureWidget(fig)


from .standard import StandardLabels

class SimpleLabels(StandardLabels):
    def __init__(self,
                 gamma_threshold: float = 0.02,
                 hadron_threshold: float = 0.2):
        super(SimpleLabels, self).__init__(gamma_threshold, hadron_threshold)

        self._labels = [
            'MIP',
            'HIP',
            'shower',
            'michel',
            'diffuse',
            'invisible'
        ]

    @property
    def pion(self):
        return self.index('MIP')

    @property
    def muon(self):
        return self.index('MIP')

    @property
    def kaon(self):
        return self.index('HIP')

    @property
    def hadron(self):
        return self.index('HIP')

    @property
    def delta(self):
        return self.index('MIP')


def ccqe(part):
    # get primary for each particle
    part = part.set_index("g4_id", drop=False)

    # convert from PDG code to label
    def label(pdg):
        if abs(pdg) == 11: return 0 # electron
        if abs(pdg) == 13: return 1 # muon
        return 2 # hadronic

    # trace lineage back from particle to primary and get label
    def func(row):
        gid = row.g4_id
        pid = row.parent_id
        while True:
            if pid == 0: return label(part.type[gid])
            # if not pid in part.g4_id: return label(part.type[gid])
            # gid = part.g4_id[pid]
            try:
                gid = part.g4_id[pid]
            except KeyError:
                return 2
            pid = part.parent_id[pid]

    # apply backtrace function to get labels
    part["semantic_label"] = part.apply(func, axis=1)
    return part.reset_index(drop=True)[["g4_id", "semantic_label"]]

def panoptic_label(part):
    part = semantic_label(part)
    part["instance_label"] = -1
    return part

def edge_label(edge):

    # False
    edge["label"] = 0

    # EM shower
    mask_e = (edge.label_1 == 0) & (edge.label_2 == 0)
    edge.loc[mask_e, "label"] = 1

    # Muon
    mask_part = (edge.g4_id_1 == edge.g4_id_2)
    mask_mu = (edge.label_1 == 1) & (edge.label_2 == 1)
    edge.loc[mask_part & mask_mu, "label"] = 2

    # Hadronic
    mask_had = (edge.label_1 == 2) & (edge.label_2 == 2)
    edge.loc[mask_part & mask_had, "label"] = 3

    return edge



from typing import Any, Callable
import numpy as np
import pandas as pd

import torch
import torch_geometric as pyg

from .base import ProcessorBase

class HitGraphProducer(ProcessorBase):
    '''Process event into graphs'''

    def __init__(self,
                 file: 'pynuml.io.File',
                 semantic_labeller: Callable = None,
                 event_labeller: Callable = None,
                 label_vertex: bool = False,
                 label_position: bool = False,
                 planes: list[str] = ['u','v','y'],
                 node_pos: list[str] = ['local_wire','local_time'],
                 pos_norm: list[float] = [0.3,0.055],
                 node_feats: list[str] = ['integral','rms'],
                 lower_bound: int = 20,
                 store_detailed_truth: bool = False):

        self.semantic_labeller = semantic_labeller
        self.event_labeller = event_labeller
        self.label_vertex = label_vertex
        self.label_position = label_position
        self.planes = planes
        self.node_pos = node_pos
        self.pos_norm = torch.tensor(pos_norm).float()
        self.node_feats = node_feats
        self.lower_bound = lower_bound
        self.store_detailed_truth = store_detailed_truth

        self.transform = pyg.transforms.Compose((
            pyg.transforms.Delaunay(),
            pyg.transforms.FaceToEdge()))

        super().__init__(file)

    @property
    def columns(self) -> dict[str, list[str]]:
        groups = {
            'hit_table': ['hit_id','local_plane','local_time','local_wire','integral','rms'],
            'spacepoint_table': ['spacepoint_id','hit_id']
        }
        if self.semantic_labeller:
            groups['particle_table'] = ['g4_id','parent_id','type','momentum','start_process','end_process']
            groups['edep_table'] = []
        if self.event_labeller:
            groups['event_table'] = ['is_cc', 'nu_pdg']
        if self.label_vertex:
            keys = ['nu_vtx_corr','nu_vtx_wire_pos','nu_vtx_wire_time']
            if 'event_table' in groups:
                groups['event_table'].extend(keys)
            else:
                groups['event_table'] = keys
        if self.label_position:
            groups["edep_table"] = []
        return groups

    @property
    def metadata(self):
        metadata = { 'planes': self.planes }
        if self.semantic_labeller is not None:
            metadata['semantic_classes'] = self.semantic_labeller.labels[:-1]
        if self.event_labeller is not None:
            metadata['event_classes'] = self.event_labeller.labels
        return metadata

    def __call__(self, evt: 'pynuml.io.Event') -> tuple[str, Any]:

        if self.event_labeller or self.label_vertex:
            event = evt['event_table'].squeeze()

        hits = evt['hit_table']
        spacepoints = evt['spacepoint_table'].reset_index(drop=True)

        # discard any events with pathologically large hit integrals
        # this is a hotfix that should be removed once the dataset is fixed
        if hits.integral.max() > 1e6:
            print('found event with pathologically large hit integral, skipping')
            return evt.name, None

        # handle energy depositions
        if self.semantic_labeller:
            edeps = evt['edep_table']
            energy_col = 'energy' if 'energy' in edeps.columns else 'energy_fraction' # for backwards compatibility
        
            # get ID of max particle
            g4_id = edeps[[energy_col, 'g4_id', 'hit_id']]
            g4_id = g4_id.sort_values(by=[energy_col],
                                      ascending=False,
                                      kind='mergesort').drop_duplicates('hit_id')
            hits = g4_id.merge(hits, on='hit_id', how='right')
            
            # charge-weighted average of 3D position
            if self.label_position:
                edeps = edeps[["hit_id", "energy", "x_position", "y_position", "z_position"]]
                for col in ["x_position", "y_position", "z_position"]:
                    edeps.loc[:, col] *= edeps.energy
                edeps = edeps.groupby("hit_id").sum()
                for col in ["x_position", "y_position", "z_position"]:
                    edeps.loc[:, col] /= edeps.energy
                edeps = edeps.drop("energy", axis="columns")
                hits = edeps.merge(hits, on="hit_id", how="right")

            hits['filter_label'] = ~hits[energy_col].isnull()
            hits = hits.drop(energy_col, axis='columns')

        # reset spacepoint index
        spacepoints = spacepoints.reset_index(names='index_3d')

        # skip events with fewer than lower_bnd simulated hits in any plane.
        # note that we can't just do a pandas groupby here, because that will
        # skip over any planes with zero hits
        for i in range(len(self.planes)):
            planehits = hits[hits.local_plane==i]
            nhits = planehits.filter_label.sum() if self.semantic_labeller else planehits.shape[0]
            if nhits < self.lower_bound:
                return evt.name, None

        # get labels for each particle
        if self.semantic_labeller:
            particles = self.semantic_labeller(evt['particle_table'])
            try:
                hits = hits.merge(particles, on='g4_id', how='left')
            except:
                print('exception occurred when merging hits and particles')
                print('hit table:', hits)
                print('particle table:', particles)
                print('skipping this event')
                return evt.name, None
            mask = (~hits.g4_id.isnull()) & (hits.semantic_label.isnull())
            if mask.any():
                print(f'found {mask.sum()} orphaned hits.')
                return evt.name, None
            del mask

        data = pyg.data.HeteroData()

        # event metadata
        r, sr, e = evt.event_id
        data['metadata'].run = r
        data['metadata'].subrun = sr
        data['metadata'].event = e

        # spacepoint nodes
        data['sp'].num_nodes = spacepoints.shape[0]

        # draw graph edges
        for i, plane_hits in hits.groupby('local_plane'):

            p = self.planes[i]
            plane_hits = plane_hits.reset_index(drop=True).reset_index(names='index_2d')

            # node position
            pos = torch.tensor(plane_hits[self.node_pos].values).float()
            data[p].pos = pos * self.pos_norm[None,:]

            # node features
            data[p].x = torch.tensor(plane_hits[self.node_feats].values).float()

            # node true position
            if self.label_position:
                data[p].c = torch.tensor(plane_hits[["x_position", "y_position", "z_position"]].values).float()

            # hit indices
            data[p].id = torch.tensor(plane_hits['hit_id'].values).long()

            # 2D edges
            data[p, 'plane', p].edge_index = self.transform(data[p]).edge_index

            # 3D edges
            edge3d = spacepoints.merge(plane_hits[['hit_id','index_2d']].add_suffix(f'_{p}'),
                                       on=f'hit_id_{p}',
                                       how='inner')
            edge3d = edge3d[[f'index_2d_{p}','index_3d']].values.transpose()
            edge3d = torch.tensor(edge3d) if edge3d.size else torch.empty((2,0))
            data[p, 'nexus', 'sp'].edge_index = edge3d.long()

            # truth information
            if self.semantic_labeller:
                data[p].y_semantic = torch.tensor(plane_hits['semantic_label'].fillna(-1).values).long()
                data[p].y_instance = torch.tensor(plane_hits['instance_label'].fillna(-1).values).long()
                if self.store_detailed_truth:
                    data[p].g4_id = torch.tensor(plane_hits['g4_id'].fillna(-1).values).long()
                    data[p].parent_id = torch.tensor(plane_hits['parent_id'].fillna(-1).values).long()
                    data[p].pdg = torch.tensor(plane_hits['type'].fillna(-1).values).long()
            if self.label_vertex:
                vtx_2d = torch.tensor([ event[f'nu_vtx_wire_pos_{i}'], event.nu_vtx_wire_time ]).float()
                data[p].y_vtx = vtx_2d * self.pos_norm[None,:]

        # event label
        if self.event_labeller:
            data['evt'].y = torch.tensor(self.event_labeller(event)).long()

        # 3D vertex truth
        if self.label_vertex:
            vtx_3d = [ [ event.nu_vtx_corr_x, event.nu_vtx_corr_y, event.nu_vtx_corr_z ] ]
            data['evt'].y_vtx = torch.tensor(vtx_3d).float()

        return evt.name, data


import os
import sys
from typing import Any

import h5py
from mpi4py import MPI


class PTOut:
    def __init__(self, outdir: str):
        self.outdir = outdir
        isExist = os.path.exists(outdir)
        if not isExist:
            rank = MPI.COMM_WORLD.Get_rank()
            if rank == 0:
                print("Error: output directory does not exist", outdir)
            sys.stdout.flush()
            MPI.COMM_WORLD.Abort(1)

    def __call__(self, name: str, obj: Any) -> None:
        import torch
        torch.save(obj, os.path.join(self.outdir, name)+".pt")

    def write_metadata(metadata: dict[str, Any]) -> None:
        raise NotImplementedError

    def exists(self, name: str) -> bool:
        return os.path.exists(os.path.join(self.outdir, name)+".pt")

class H5Out:
    def __init__(self, fname: str, overwrite: bool = False):
        # This implements one-file-per-process I/O strategy.
        # append MPI process rank to the output file name
        rank = MPI.COMM_WORLD.Get_rank()
        file_ext = ".{:04d}.h5"
        self.fname = fname + file_ext.format(rank)
        if os.path.exists(self.fname):
            if overwrite:
                os.remove(self.fname)
            else:
                print(f"Error: file already exists: {self.fname}")
                sys.stdout.flush()
                MPI.COMM_WORLD.Abort(1)
        # open/create the HDF5 file
        self.f = h5py.File(self.fname, "w")

        from .h5interface import H5Interface
        self.interface = H5Interface(self.f)
        # print(f"{rank}: creating {self.fname}")
        # sys.stdout.flush()

    def __call__(self, name: str, obj: Any) -> None:
        """
        for key, val in obj:
            # set chunk sizes to val shape, so there is only one chunk per dataset
            # if isinstance(val, torch.Tensor) and val.nelement() == 0 :
            #   print("zero val ",name,"/",key," shape=",val.shape)
            if isinstance(val, torch.Tensor) and val.nelement() > 0 :
                # Note compressed datasets can only be read/written in MPI collective I/O mode in HDF5
                self.f.create_dataset(f"/{name}/{key}", data=val, chunks=val.shape, compression="gzip")
                # The line below is to not enable chunking/compression
                # self.f.create_dataset(f"/{name}/{key}", data=val)
            else:
                # if data is not a tensor or is empty, then disable chunking/compression
                self.f.create_dataset(f"/{name}/{key}", data=val)
        """
        import numpy as np
        import torch_geometric as pyg

        # collect and construct fields of compound data type
        fields = []
        data = ()

        # special treatment for heterograph object
        if isinstance(obj, pyg.data.HeteroData):
            self.interface.save(name, obj)
            return
        for key, val in obj:
            if np.isscalar(val): # only n_sp is a scalar
                data = data + (val,)
                field = (key, type(val))
            else:
                if val.nelement() == 0: # save tensor with zero-sized dimension as a scalar 0
                    # HDF5 compound data type does not allow zero-size dimension
                    # ValueError: Zero-sized dimension specified (zero-sized dimension specified)
                    val = val.numpy()  # convert a tensor to numpy
                    data = data + (0,)
                    field = (key, val.dtype)
                else:
                    val = val.numpy() # convert a tensor to numpy
                    data = data + (val,)
                    field = (key, val.dtype, val.shape)
            fields.append(field)
        ctype = np.dtype(fields)
        # create a scalar dataset of compound data type
        ds = self.f.create_dataset(f"/{name}", shape=(), dtype=ctype, data=data)
        del ctype, fields, data, ds

    def write_metadata(self, metadata: dict[str, Any]) -> None:
        for key, val in metadata.items():
            self.f[key] = val

    def __del__(self):
        if self.f != None: self.f.close()


import pandas as pd

class FlavorLabels:
    def __init__(self):
        self._labels = (
            'cc_nue',
            'cc_numu',
            'cc_nutau',
            'nc')

    @property
    def labels(self):
        return self._labels

    def label(self, idx: int):
        if not 0 <= label < len(self._labels):
            raise Exception(f'index {idx} out of range for {len(self._labels)} labels.')
        return self._labels[idx]

    def index(self, name: str):
        if name not in self._labels:
            raise Exception(f'"{name}" is not the name of a class.')
        return self._labels.index(name)

    @property
    def cc_nue(self):
        return self.index('cc_nue')
    
    @property
    def cc_numu(self):
        return self.index('cc_numu')
    
    @property
    def cc_nutau(self):
        return self.index('cc_nutau')
    
    @property
    def nc(self):
        return self.index('nc')

    def __call__(self, event: pd.Series):
        if not event.is_cc:
            return self.nc
        pdg = abs(event.nu_pdg)
        if pdg == 12:
            return self.cc_nue
        if pdg == 14:
            return self.cc_numu
        if pdg == 16:
            return self.cc_nutau
        raise Exception(f'PDG code {event.nu_pdg} not recognised.')


from typing import Any

import h5py
import numpy as np
import torch
from torch_geometric.data import Data, HeteroData


class H5Interface:
    def __init__(self, file: h5py.File):
        self.f = file

    def save_data(self, data: Data) -> None:
        code

    def _add_dataset(self, key: str, val: Any) -> None:
        if np.isscalar(val):
            self._data = self._data + (val,)
            field = (key, type(val))
        else:
            if val.nelement() == 0: # save tensor with zero-sized dimension as a scalar 0
                # HDF5 compound data type does not allow zero-size dimension
                # ValueError: Zero-sized dimension specified (zero-sized dimension specified)
                self._data = self._data + (0,)
                field = (key, val.numpy().dtype)
            else:
                val = val.numpy() # convert a tensor to numpy
                self._data = self._data + (val,)
                field = (key, val.dtype, val.shape)
        self._fields.append(field)

    def save_heterodata(self, data: HeteroData) -> None:

        self._data = ()
        self._fields = []

        nodes, edges = data.metadata()

        # save node stores
        for node in nodes:
            if "_" in node:
                raise Exception(f'"{node}" is not a valid node store name! Underscores are not supported.')
            for key in data[node].keys():
                self._add_dataset(f'{node}/{key}', data[node][key])

        # save edge stores
        for edge in edges:
            for tmp in edge:
                if "_" in tmp:
                    raise Exception(f'"{tmp}" is not a valid edge store name component! Underscores are not supported.')
            name = "_".join(edge)
            for key in data[edge].keys():
                self._add_dataset(f'{name}/{key}', data[edge][key])

    def save(self, name: str, data: Any) -> None:
        if isinstance(data, Data):
            self.save_data(data)
        elif isinstance(data, HeteroData):
            self.save_heterodata(data)
        else:
            raise NotImplementedError(f'No save method implemented for {type(data)}!')

        # create a scalar dataset of compound data type
        ctype = np.dtype(self._fields)
        ds = self.f.create_dataset(f'/dataset/{name}', shape=(), dtype=ctype, data=self._data)
        del ctype, self._fields, self._data, ds

    def load_heterodata(self, name: str) -> HeteroData:
        data = HeteroData()
        # Read the whole dataset idx, dataset name is self.groups[idx]
        group = self.f[f'dataset/{name}'][()]
        for dataset in group.dtype.names:
            store, attr = dataset.split('/')
            if "_" in store: store = tuple(store.split("_"))
            if group[dataset].ndim == 0:
                if attr == 'edge_index': # empty edge tensor
                    data[store][attr] = torch.LongTensor([[],[]])
                else: # scalar
                    data[store][attr] = torch.as_tensor(group[dataset][()])
            else: # multi-dimension array
                data[store][attr] = torch.as_tensor(group[dataset][:])
        return data

    def keys(self) -> list[str]:
        return list(self.f['dataset'].keys())